(*
Author:  Christina Kohl <christina.kohl@uibk.ac.at> (2022)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2013-2017)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015, 2023)
Author:  Dohan Kim <dohan.kim@uibk.ac.at> (2023)

License: LGPL (see file COPYING.LESSER)
*)
theory Critical_Pair_Closure_Impl
imports
  Critical_Pairs_Impl
  Parallel_Closed
  Strongly_Closed
  Critical_Pair_Closing_Systems
  Framework.QDP_Framework_Impl
  Development_Closed
  Parallel_Critical_Pairs
begin

definition check_strongly_closed where
  "check_strongly_closed ren R n = do {
     check_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (\<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0))) \<and>
         \<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R t n) x) (reachable_terms R s (Suc 0))))
       (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
        showsl_lit (STR '' is not strongly closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl ren R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not strongly closed\<newline>'') \<circ> showsl_trs R)"

lemma check_strongly_closed:
  assumes "isOK(check_strongly_closed ren R n)"
  shows "CR (rstep (set R))"
proof (rule strongly_closed_linear_CR)
  from assms show "linear_trs (set R)" unfolding check_strongly_closed_def by auto
  let ?red = "\<lambda>s n. reachable_terms R s n"
  let ?opt = "\<lambda>s t. List.find (\<lambda>x. List.member (?red s n) x) (?red t (Suc 0))"
  { fix s t
    assume "\<not> Option.is_none (?opt s t)"
    then obtain r where r: "?opt s t = Some r" by (cases "?opt s t") auto
    then have "List.member (?red s n) r" unfolding find_Some_iff by auto
    then have mem:"r \<in> set (?red s n)" using in_set_member by fast
    from r have "\<exists>i<length (?red t (Suc 0)). r = (?red t (Suc 0)) ! i" unfolding find_Some_iff by auto
    then have "r \<in> set (?red t (Suc 0))" using nth_mem by blast
    with mem have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      using reachable_terms reachable_terms_one by metis
  } note sound=this
  { fix b s t
    assume "(b, s, t) \<in> critical_pairs ren (set R) (set R)"
    with assms have *:"\<not> Option.is_none (?opt s t) \<and> \<not> Option.is_none (?opt t s)"
      unfolding check_strongly_closed_def by fastforce
    with sound have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      and  "\<exists>u. (t, u) \<in> (rstep (set R))\<^sup>* \<and> (s, u) \<in> (rstep (set R))\<^sup>=" by blast+
  }
  then show "strongly_closed ren (set R)" unfolding strongly_closed_def by blast
qed

context 
  fixes R :: "('f,'v)rules" and V :: "'v set"  
begin
fun is_par_rstep_var_restr :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_par_rstep_var_restr (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> 
    vars_term (Fun g ts) \<inter> V = {} \<and> (Fun f ss, Fun g ts) \<in> rrstep (set R) \<or>
    (f = g \<and> length ss = length ts \<and> list_all2 is_par_rstep_var_restr ss ts))"
| "is_par_rstep_var_restr s t = (s = t \<or> vars_term t \<inter> V = {} \<and> (s,t) \<in> rrstep (set R))"

lemma is_par_rstep_code_helper: "vars_term t \<inter> V = {} \<longleftrightarrow>
  (\<forall> x \<in> set (vars_term_list t). x \<notin> V)" 
  by auto

lemmas is_par_rstep_var_restr_code[code] = is_par_rstep_var_restr.simps[unfolded is_par_rstep_code_helper]

lemma is_par_rstep_var_restr[simp]:
  "is_par_rstep_var_restr s t \<longleftrightarrow> (s, t) \<in> par_rstep_var_restr (set R) V"
proof 
  let ?Prop = "\<lambda> s t. s = t \<or> vars_term t \<inter> V = {} \<and> (s,t) \<in> rrstep (set R)" 
  {
    fix s t
    assume "?Prop s t" 
    hence "\<exists> C infos. (s, t) \<in> par_rstep_mctxt (set R) C infos \<and> vars_below_hole t C \<inter> V = {}"
    proof 
      assume "s = t" 
      thus ?thesis by (intro exI[of _ "mctxt_of_term s"] exI[of _ Nil], auto simp: par_rstep_mctxt_reflI)
    next
      assume "vars_term t \<inter> V = {} \<and> (s,t) \<in> rrstep (set R)"      
      then obtain l r \<sigma> where id: "s = l \<cdot> \<sigma>" "t = r \<cdot> \<sigma>" and 
        lr: "(l,r) \<in> set R" and
        vars: "vars_term t \<inter> V = {}" 
        by (metis rrstepE)      
      thus ?thesis by (intro exI[of _ MHole] exI[of _ "[Par_Info s t (l,r)]"], auto intro: par_rstep_mctxt_MHoleI)
    qed
  } note Prop = this  
  {
    assume "is_par_rstep_var_restr s t" 
    hence "\<exists> C infos. (s, t) \<in> par_rstep_mctxt (set R) C infos \<and> vars_below_hole t C \<inter> V = {}"
    proof (induct rule: is_par_rstep_var_restr.induct[])
      case "2_1" 
      thus ?case by (intro Prop, auto)
    next
      case "2_2" 
      thus ?case by (intro Prop, auto)
    next
      case (1 f ss g ts)
      show ?case
      proof (cases "?Prop (Fun f ss) (Fun g ts)")
        case True
        thus ?thesis using Prop by auto
      next
        case False
        with 1 have args: "f = g" "length ss = length ts" "list_all2 is_par_rstep_var_restr ss ts"
          by (auto split: if_splits)
        let ?P = "\<lambda> i C infos. (ss ! i, ts ! i) \<in> par_rstep_mctxt (set R) C infos \<and> vars_below_hole (ts ! i) C \<subseteq> (UNIV - V)" 
        { fix i
          assume i:"i < length ss"
          then have si:"ss ! i \<in> set ss" by auto
          from i args(2) have ti:"ts ! i \<in> set ts" by auto
          from args(3) have iprv:"is_par_rstep_var_restr (ss ! i) (ts ! i)" using i list_all2_nthD by blast
          with 1(1)[of "ss!i" "ts!i"] have pp:"\<exists>C infos. ?P i C infos" 
            using local.args(1) local.args(2) using si ti by blast
        } 
        hence "\<forall> i. \<exists> C infos. i < length ss \<longrightarrow> ?P i C infos" by blast
        from choice[OF this] obtain C where "\<forall> i. \<exists> infos. i < length ss \<longrightarrow> ?P i (C i) infos" by blast
        from choice[OF this] obtain infos where IH: "\<And> i. i < length ss \<Longrightarrow> ?P i (C i) (infos i)" by blast
        let ?C = "MFun f (map C [0..<length ss])" 
        let ?infos = "concat (map infos [0..<length ss])" 
        show ?thesis 
        proof (intro exI[of _ ?C] exI[of _ ?infos] conjI)
          show "vars_below_hole (Fun g ts) ?C \<inter> V = {}" using IH args(2) unfolding args(1)
            by (subst vars_below_hole_Fun; force)
          show "(Fun f ss, Fun g ts) \<in> par_rstep_mctxt (set R) ?C ?infos" unfolding args(1) using args(2) IH
            by (intro par_rstep_mctxt_funI, auto)
        qed
      qed
    qed
    thus " (s, t) \<in> par_rstep_var_restr (set R) V" unfolding par_rstep_var_restr_def by auto
  }
  {
    assume "(s, t) \<in> par_rstep_var_restr (set R) V" 
    from this[unfolded par_rstep_var_restr_def] obtain C infos where
       st: "(s, t) \<in> par_rstep_mctxt (set R) C infos" and vars: "vars_below_hole t C \<inter> V = {}" by auto
    thus "is_par_rstep_var_restr s t" 
    proof (induct C arbitrary: s t infos)
      case (MVar x)
      from par_rstep_mctxt_MVarE[OF MVar(1)]
      have "s = Var x" "t = Var x" by auto
      thus ?case by simp
    next
      case MHole
      from par_rstep_mctxt_MHoleE[OF MHole(1)]
      have "(s,t) \<in> rrstep (set R)" by auto  
      then show ?case using MHole(2) by (cases s; cases t; auto)
    next
      case (MFun f Cs)
      from par_rstep_mctxt_MFunD[OF MFun(2)]
      obtain ss ts Infos
        where s: "s = Fun f ss" and
          t: "t = Fun f ts" and 
          len: "length ss = length Cs" 
            "length ts = length Cs"
            "length Infos = length Cs" and
          infos: "infos = concat Infos" and
          steps: "\<And> i. i < length Cs \<Longrightarrow>(ss ! i, ts ! i) \<in> par_rstep_mctxt (set R) (Cs ! i) (Infos ! i)"
        by auto
      show ?case unfolding s t is_par_rstep_var_restr.simps
      proof (intro disjI2 conjI refl list_all2_all_nthI, unfold len)
        fix i
        assume i: "i < length Cs" 
        hence mem: "Cs ! i \<in> set Cs" by auto
        show "is_par_rstep_var_restr (ss ! i) (ts ! i)" 
        proof (rule MFun(1)[OF mem steps[OF i]])
          have "vars_below_hole (ts ! i) (Cs ! i) \<subseteq> vars_below_hole t (MFun f Cs)" 
            unfolding t using i len 
            by (subst vars_below_hole_Fun, auto)
          with MFun(3) show "vars_below_hole (ts ! i) (Cs ! i) \<inter> V = {}" by auto
        qed
      qed auto
    qed
  }
qed
end

lemma par_rstep_var_restr_code[code_unfold]: 
  "(s, t) \<in> par_rstep_var_restr (set R) V \<longleftrightarrow> is_par_rstep_var_restr R V s t" 
  by simp

fun is_par_rstep :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_par_rstep R (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> (Fun f ss, Fun g ts) \<in> rrstep (set R) \<or>
    (f = g \<and> length ss = length ts \<and> list_all2 (is_par_rstep R) ss ts))"
| "is_par_rstep R s t = (s = t \<or> (s,t) \<in> rrstep (set R))"

lemma is_par_rstep[simp]:
  "is_par_rstep R s t \<longleftrightarrow> (s, t) \<in> par_rstep (set R)"
proof -
  have "is_par_rstep R s t = is_par_rstep_var_restr R {} s t"
    by (induct R s t rule: is_par_rstep.induct, auto simp del: is_par_rstep_var_restr simp: list_all2_conv_all_nth)
  also have "\<dots> \<longleftrightarrow> (s, t) \<in> par_rstep_var_restr (set R) {}" by simp
  also have "\<dots> \<longleftrightarrow> (s, t) \<in> par_rstep (set R)" 
    unfolding par_rstep_var_restr_def par_rstep_par_rstep_mctxt_conv by auto
  finally show ?thesis .
qed

lemma par_rstep_code[code_unfold]: "(s, t) \<in> par_rstep (set R) \<longleftrightarrow> is_par_rstep R s t" by simp

fun root_rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "root_rewrite R s = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [(r \<cdot> \<sigma>)])) R)"

lemma root_rewrite_sound:
  assumes "t \<in> set (root_rewrite R s)"
  shows "(s, t) \<in> rrstep (set R)"
proof -
  from assms
  have "\<exists> l r.  (l,r) \<in> set R \<and> t \<in> set (case match s l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [r \<cdot> \<sigma>])"
    by auto
  from this obtain l r where one:
    "(l,r) \<in> set R \<and> t \<in> set (case match s l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [r \<cdot> \<sigma>])"
    by auto
  from this obtain \<sigma> where two: "match s l = Some \<sigma> \<and> t \<in> {r \<cdot> \<sigma>}" by (cases "match s l", auto)
  then have match: "l \<cdot> \<sigma> = s" using match_sound by auto
  with one match one two have "(s,t) \<in> rstep_r_p_s (set R) (l,r) [] \<sigma>" unfolding rstep_r_p_s_def by (simp add: Let_def ctxt_supt_id)
  then show "(s,t) \<in> rrstep (set R)" unfolding rstep_iff_rstep_r_p_s rrstep_def by blast
qed

fun parallel_rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "parallel_rewrite R (Var x) = [Var x]"
| "parallel_rewrite R (Fun f ss) = remdups
     (root_rewrite R (Fun f ss) @ map (\<lambda>ss. Fun f ss) (product_lists (map (parallel_rewrite R) ss)))"

lemma parallel_rewrite_par_step:
  assumes "t \<in> set (parallel_rewrite R s)"
  shows "(s, t) \<in> par_rstep (set R)"
using assms
proof (induct s arbitrary: t)
  case (Fun f ss)
  then consider (root) "t \<in> set (root_rewrite R (Fun f ss))"
    | (args) "t \<in> set (map (\<lambda>ss. Fun f ss) (product_lists (map (parallel_rewrite R) ss)))"
    by force
  then show ?case
  proof (cases)
    case root
    from root_rewrite_sound[OF this] obtain l r \<sigma> where "(l, r) \<in> set R"
      and "l \<cdot> \<sigma> = Fun f ss" and "r \<cdot> \<sigma> = t"
      unfolding rrstep_def rstep_r_p_s_def by auto
    then show ?thesis using par_rstep.intros(1) by metis
  next
    case args
    then obtain ts where t:"t = Fun f ts" and ts:"ts \<in> set (product_lists (map (parallel_rewrite R) ss))"
      by auto
    then have len:"length ss = length ts" using in_set_product_lists_length by force
    { fix i
      assume i:"i < length ts"
      have "ts ! i \<in> set (parallel_rewrite R (ss ! i))"
        using ts[unfolded product_lists_set[of "_ ss"]]
        by (auto simp: list_all2_map2[of "(\<lambda>x ys. x \<in> set ys)"] intro: list_all2_nthD[OF _ i])
      with Fun(1) len i have "(ss ! i, ts ! i) \<in> par_rstep (set R)" by auto
    }
    from par_rstep.intros(2)[OF this len] show ?thesis using t by auto
  qed
qed auto

(*Takes a list of rewrite rules and two terms and checks for each rule whether the first term matches
the lhs and the second term the rhs of the rule. For each rule where this is the case,
returns the lists of terms corresponding to the matching substitutions. 
Terms that correspond to variables that only appear in the lhs of the rule (i.e. terms that are being
erased) don't play a role in the remainder of the computation for multisteps hence they can be ignored.*) 
fun root_steps_substs :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f, 'v) term list \<times> ('f, 'v) term list) list"
where
  "root_steps_substs R s t = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> (case match t r of 
        None \<Rightarrow> []
      | Some \<tau> \<Rightarrow> [(map \<sigma> (var r), map \<tau> (var r))]))) 
  R)"

lemma root_steps_substs_exists:
  assumes "(ss, ts) \<in> set (root_steps_substs R s t)"
  shows "\<exists> l r \<sigma> \<tau>. (l, r) \<in> set R \<and> l \<cdot> \<sigma> = s \<and> r \<cdot> \<tau> = t \<and> (ss, ts) = (map \<sigma> (var r), map \<tau> (var r))"
proof-
from assms obtain l r where lr:"(l,r) \<in> set R" "(ss, ts) \<in> set (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> (case match t r of 
        None \<Rightarrow> []
      | Some \<tau> \<Rightarrow> [(map \<sigma> (var r), map \<tau> (var r))]))"
  by auto
  then obtain \<sigma> where \<sigma>:"match s l = Some \<sigma>"
    by fastforce 
  with lr obtain \<tau> where \<tau>:"match t r = Some \<tau>" 
    by fastforce
  from lr \<sigma> \<tau> have "(ss, ts) = (map \<sigma> (var r), map \<tau> (var r))"
    by simp 
  with lr(1) \<sigma> \<tau> show ?thesis
    using match_matches by blast
qed

lemma size_match_subst_Fun: (*used for termination_simps*)
  assumes "is_Fun l" and "x \<in> vars_term l"
  and match:"match s l = Some \<tau>"
shows "size (\<tau> x) < size s"
proof-
  from assms(1) obtain f ts where l:"l = Fun f ts"
    by blast 
  from match have *:"l \<cdot> \<tau> = s"
    by (simp add: match_matches) 
  then obtain ss where s:"s = Fun f ss"
    unfolding l by force
  from assms(2) obtain i where i:"i < length ts" and x:"x \<in> vars_term (ts!i)" 
    unfolding l by (metis term.sel(4) var_imp_var_of_arg) 
  from * have le:"length ts = length ss" 
    unfolding s l by auto 
  moreover from * i l s have "ts!i \<cdot> \<tau> = ss!i" 
    by fastforce
  then have "size (\<tau> x) \<le> size (ss!i)"
    using match_subst_vars_term_size x by metis
  with i show ?thesis unfolding s term.size le
    by (metis add.commute add_0 add_Suc in_set_conv_nth less_Suc_eq_le size_list_estimation')
qed

fun root_subst_with_rhs :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f, 'v) term \<times> ('f, 'v) term list) list"
where
  "root_subst_with_rhs R s = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [(r, map \<sigma> (var r))]))
  R)"

lemma root_steps_subst_rhs_exists:
  assumes "(r, ss) \<in> set (root_subst_with_rhs R s)"
  shows "\<exists> l \<sigma>. (l, r) \<in> set R \<and> l \<cdot> \<sigma> = s \<and> ss = map \<sigma> (var r)"
proof-
from assms obtain l where lr:"(l,r) \<in> set R" "(r, ss) \<in> set (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [(r, map \<sigma> (var r))])"
  by auto
  then obtain \<sigma> where \<sigma>:"match s l = Some \<sigma>"
    by fastforce
  with lr show ?thesis
    using match_matches by force
qed

context
  fixes R :: "('f, 'v) rules"
  assumes *:"list_all (\<lambda>(l, r). is_Fun l \<and> (vars_term r \<subseteq> vars_term l)) R"
begin

lemma varcond:
  "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
  using * Ball_set_list_all case_prodD by (metis (no_types, lifting)) 

lemma [termination_simp]:
  assumes "(l, r) \<in> set R" 
    and "Some \<sigma> = match (Fun g ts) l" 
    and "x \<in> vars_term r"
  shows "size (\<sigma> x) < Suc (size_list size ts)"
  using assms size_match_subst_Fun varcond
  by (metis (no_types, lifting) add.right_neutral add_Suc_right subsetD term.size(4))

lemma [termination_simp]:
  assumes "(l, r) \<in> set R"
    and "Some \<tau> = match s l" 
    and "x \<in> vars_term r" 
  shows "size (\<tau> x) < size s"
  using assms size_match_subst_Fun varcond by (metis subset_iff)

lemma [termination_simp]:
  assumes "(l, r) \<in> set R"
    and "Some \<tau> = match (Var v) l" 
  shows "size (\<tau> x) = 0"
proof-
  from assms(1) varcond obtain f ss where "l = Fun f ss"
    by blast 
  with assms(2) have False
    by (metis match_matches subst_apply_eq_Var term.distinct(1)) 
  then show ?thesis 
    by simp
qed

text\<open>Given two terms s and t determine whether there exists a multi-step from s to t in the given TRS.\<close>
(*Only terminating if variable left-hand sides are excluded.*)
fun is_mstep :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
  "is_mstep (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> 
    list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (Fun f ss) (Fun g ts)) \<or> 
    (if f = g \<and> length ss = length ts then list_all2 is_mstep ss ts else False))" 
| "is_mstep s t = (s = t \<or> list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R s t))"

text\<open>Show that all multisteps are covered by the definition above.\<close>
(*This is not necessary for compiling Ceta but good to know - kind of a sanity check.*)
lemma mstep_is_mstep:
  assumes "(s, t) \<in> mstep (set R)"
  shows "is_mstep s t"
  using assms proof(induct)
  case (args f n ss ts)
  then have "list_all2 is_mstep ss ts"
    by (simp add: list_all2_all_nthI) 
  with args show ?case
    by simp 
next
  case (rule l r \<sigma> \<tau>)
  from rule(1) obtain i where i:"i < length R" "R!i = (l, r)"
    by (metis in_set_idx) 
  obtain \<sigma>' where sigma:"match (l \<cdot> \<sigma>) l = Some \<sigma>'" "(\<forall>x\<in>vars_term l. \<sigma> x = \<sigma>' x)" 
    by (meson match_complete')
  obtain \<tau>' where tau:"match (r \<cdot> \<tau>) r = Some \<tau>'" "(\<forall>x\<in>vars_term r. \<tau> x = \<tau>' x)" 
    by (meson match_complete')
  let ?matches="(map (\<lambda>(l', r'). case match (l \<cdot> \<sigma>) l' of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> case match (r \<cdot> \<tau>) r' of None \<Rightarrow> [] | Some \<tau> \<Rightarrow> [(map \<sigma> (var r'), map \<tau> (var r'))]) R)"
  have "i < length ?matches"
    using i(1) by auto 
  moreover have "(map \<sigma>' (var r), map \<tau>' (var r)) \<in> set (?matches ! i)"
    using sigma(1) tau(1) i by simp
  ultimately have "(map \<sigma>' (var r), map \<tau>' (var r)) \<in> set (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))"
    unfolding root_steps_substs.simps by (metis (no_types, lifting) concat_nth concat_nth_length in_set_conv_nth) 
  then obtain j where j:"j < length (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))" "root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>) ! j = (map \<sigma>' (var r), map \<tau>' (var r))"
    by (metis in_set_idx)
  have "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))!j)"
  proof-
    {fix i assume "i < length (var r)"
      then have *:"var r!i \<in> vars_term r"
        by (metis comp_apply nth_mem set_remdups set_rev set_vars_term_list) 
      then have "var r!i \<in> vars_term l" 
        using varcond rule.hyps(1) by blast 
      moreover then have "\<sigma>' (var r !i) = \<sigma> (var r ! i)" 
        using sigma(2) * by simp
      moreover have "\<tau>' (var r !i) = \<tau> (var r ! i)"
        using * tau(2) by simp
      ultimately have "is_mstep (\<sigma>' (var r !i)) (\<tau>' (var r!i))" 
        using rule(2) by force
    }
    then have "list_all2 is_mstep (map \<sigma>' (var r)) (map \<tau>' (var r))"
      by (simp add: list_all2_conv_all_nth) 
    then show ?thesis 
      unfolding j(2) by fastforce 
  qed
  then have *:"list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))"
    using j by (meson list_ex_length) 
  then show ?case proof(cases "l \<cdot> \<sigma>")
    case (Var x)
    then show ?thesis 
      using * is_mstep.simps(2) by presburger 
  next
    case (Fun f ts)
    note f=this
    show ?thesis proof(cases "r \<cdot> \<tau>")
      case (Var x)
      then show ?thesis 
        using * is_mstep.simps(3) by presburger 
    next
      case (Fun g ss)
      then show ?thesis 
        using f * is_mstep.simps(1) by presburger 
    qed
  qed 
qed simp

lemma is_mstep:
  assumes "is_mstep s t"
  shows "(s, t) \<in> mstep (set R)"
using assms proof (induct rule: is_mstep.induct)
  case (1 f ss g ts)
  then consider (eq) "Fun f ss = Fun g ts"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (Fun f ss) (Fun g ts))"
    | (args) "f = g" and "length ss = length ts" and "list_all2 is_mstep ss ts"
    by (auto split: if_splits)
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R (Fun f ss) (Fun g ts))" "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R (Fun f ss) (Fun g ts))!i)"
      using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R (Fun f ss) (Fun g ts))" "list_all2 is_mstep ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = Fun f ss" and r:"r \<cdot> \<tau> = Fun g ts" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = Fun g ts"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau> by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with 1(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis 
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  next
    case args
    { fix i
      assume i:"i < length ss"
      then have si:"ss ! i \<in> set ss" by auto
      from i args(2) have ti:"ts ! i \<in> set ts" by auto
      from args(3) have "is_mstep (ss ! i) (ts ! i)" using i list_all2_nthD by blast
      with 1(2)[of "ss ! i" "ts ! i"] args(1,2) si ti have "(ss ! i, ts ! i) \<in> mstep (set R)"
        by auto
    }
    then show ?thesis using args(1,2)
      by (simp add: mstep.args)
  qed simp
next
  case ("2_1" v t) 
  then consider (eq) "Var v = t"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R (Var v) t)"
    by auto
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R (Var v) t)" "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R (Var v) t)!i)"
      unfolding is_mstep.simps using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R (Var v) t)" "list_all2 is_mstep ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = Var v" and r:"r \<cdot> \<tau> = t" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = t"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau>  by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with "2_1"(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  qed simp
next
  case ("2_2" s v)
  then consider (eq) "s = Var v"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep ss ts) (root_steps_substs R s (Var v))"
    by auto
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R s (Var v))" "(\<lambda> (ss, ts). list_all2 is_mstep ss ts) ((root_steps_substs R s (Var v))!i)"
      unfolding is_mstep.simps using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R s (Var v))" "list_all2 is_mstep ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = s" and r:"r \<cdot> \<tau> = Var v" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = Var v"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau>  by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with "2_2"(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  qed simp
qed

(*compute list of terms reachable in multi-step from a given term.*)
fun mstep_rewrite :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "mstep_rewrite (Var x) = [Var x]"
| "mstep_rewrite (Fun f ss) = remdups (
     (concat (map (\<lambda>(r, ts). 
        (map (\<lambda>args. r \<cdot> (mk_subst Var (zip (var r) args))) (product_lists (map mstep_rewrite ts))))
      (root_subst_with_rhs R (Fun f ss))))
    @(map (\<lambda>ss. Fun f ss) (product_lists (map mstep_rewrite ss))))"

lemma mstep_rewrite_mstep:
  assumes "t \<in> set (mstep_rewrite s)"
  shows "(s, t) \<in> mstep (set R)"
using assms
proof (induct s arbitrary: t rule:subterm_induct)
  case (subterm s)
  then show ?case proof(cases s)
    case (Var x)
    with subterm(2) show ?thesis by simp
  next
    case (Fun f ss)
    with subterm consider (root) "t \<in> set (concat (map (\<lambda>(r,ts).(map (\<lambda>args. r \<cdot> (mk_subst Var (zip (var r) args))) 
          (product_lists (map mstep_rewrite ts)))) (root_subst_with_rhs R (Fun f ss))))"
      | (args) "t \<in> set (map (\<lambda>ss. Fun f ss) (product_lists (map mstep_rewrite ss)))"
      by force 
    then show ?thesis
    proof (cases)
      case root
      then obtain r ts where rhs_subst:"(r,ts) \<in> set (root_subst_with_rhs R (Fun f ss))" 
                             "t \<in> set (map (\<lambda>args. r \<cdot> (mk_subst Var (zip (var r) args))) (product_lists (map mstep_rewrite ts)))"
        by force
      from root_steps_subst_rhs_exists[OF rhs_subst(1)] obtain l \<sigma> where lr:"(l, r) \<in> set R"
        and sigma:"l \<cdot> \<sigma> = Fun f ss" "ts = map \<sigma> (var r)" by auto
      from rhs_subst(2) obtain args where args:"t = r \<cdot> (mk_subst Var (zip (var r) args))"
                                      "args \<in> set (product_lists (map mstep_rewrite ts))"
        by auto    
      then have len:"length args = length ts"
        using in_set_product_lists_length by fastforce 
      then have len':"length args = length (var r)"
        by (simp add: sigma(2)) 
      let ?\<tau>="\<lambda>x. if x \<in> vars_term r then (mk_subst Var (zip (var r) args)) x else \<sigma> x"
      from args(1) have t:"t = r \<cdot> ?\<tau>"
        by (simp add: term_subst_eq_conv) 
      { fix x
        assume x:"x \<in> vars_term l"
        have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
          case True
          then obtain i where i:"i < length (var r)" "x = var r ! i"
            by (metis in_set_idx set_vars_term_list vars_term_list_var)
          with True len' have tau_x:"?\<tau> x = args!i"
            by (simp add: mk_subst_distinct)
          from i sigma(2) have sigma_x:"\<sigma> x = ts!i"
            by simp 
          have "\<sigma> x \<lhd> Fun f ss"
            by (metis is_VarI lr sigma(1) subst_image_subterm term.set_cases(2) varcond x) 
          with sigma_x have "ts!i \<lhd> Fun f ss" by simp
          moreover have "args!i \<in> set (mstep_rewrite (ts!i))" using args(2) i(1) len' len 
            unfolding product_lists_set list_all2_conv_all_nth by force
          ultimately show ?thesis using subterm(1) sigma_x tau_x unfolding Fun by presburger
        next
          case False 
          then show ?thesis by simp
        qed
      }
      then show ?thesis using mstep.intros(3)[OF lr] sigma(1) unfolding Fun t
        by fastforce
    next
      case args
      then obtain ts where t:"t = Fun f ts" and ts:"ts \<in> set (product_lists (map mstep_rewrite ss))"
        by auto
      then have len:"length ss = length ts" using in_set_product_lists_length by force
      { fix i
        assume i:"i < length ts"
        have "ts ! i \<in> set (mstep_rewrite (ss ! i))"
          using ts[unfolded product_lists_set[of "_ ss"]]
          by (auto simp: list_all2_map2[of "(\<lambda>x ys. x \<in> set ys)"] intro: list_all2_nthD[OF _ i])
        with subterm len i have "(ss ! i, ts ! i) \<in> mstep (set R)" 
          unfolding Fun by auto
      }
      with mstep.intros(2) len t Fun show ?thesis
        by metis
    qed
  qed
qed

end

text\<open>We need to be able to export code for @{const is_mstep} and @{const mstep_rewrite}, hence the following definitions
(adapted from template by Rene).\<close>
(*New type for well-formed TRSs*)
typedef ('f, 'v) wfTRS = "{R :: ('f, 'v) rules. list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R}"
  using list_all_simps(2) by blast 

setup_lifting type_definition_wfTRS

lift_definition get_TRS :: "('f, 'v) wfTRS \<Rightarrow> ('f, 'v) rules" is "\<lambda> R. R" .

lemma is_wf_get_TRS: "list_all (\<lambda>(l,r). is_Fun l  \<and> vars_term r \<subseteq> vars_term l) (get_TRS R')" 
  by (transfer, auto)

definition "is_mstep' R s t = is_mstep (get_TRS R) s t" 
definition "mstep_rewrite' R s = mstep_rewrite (get_TRS R) s" 

lemmas is_mstep'_simps = is_mstep.simps[OF is_wf_get_TRS, folded is_mstep'_def]
declare is_mstep'_simps[code]

lemmas mstep_rewrite'_simps = mstep_rewrite.simps[OF is_wf_get_TRS, folded mstep_rewrite'_def]
declare mstep_rewrite'_simps[code]

lift_definition (code_dt) get_wfTRS :: "('f, 'v) rules \<Rightarrow> ('f, 'v) wfTRS option" is 
  "\<lambda> R. if list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R then Some R else None"
  by auto

definition err_wf where "err_wf = STR ''TRS is not well-formed''" 
(*should actually never be printed, since TRS is checked before even calling is_mstep*) 

lemma [code]: "is_mstep R = (case get_wfTRS R of Some R' \<Rightarrow> is_mstep' R' 
  | None \<Rightarrow> Code.abort err_wf (\<lambda> _. is_mstep R ))" 
proof (cases "list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R")
  case False
  hence "get_wfTRS R = None" by transfer auto
  thus ?thesis by auto
next
  case True
  hence "\<exists> R'. get_wfTRS R = Some R' \<and> get_TRS R' = R"
    by (transfer, auto)
  thus ?thesis unfolding is_mstep'_def by auto
qed

lemma [code]: "mstep_rewrite R = (case get_wfTRS R of Some R' \<Rightarrow> mstep_rewrite' R' 
  | None \<Rightarrow> Code.abort err_wf (\<lambda> _. mstep_rewrite R ))" 
proof (cases "list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R")
  case False
  hence "get_wfTRS R = None" by transfer auto
  thus ?thesis by auto
next
  case True
  hence "\<exists> R'. get_wfTRS R = Some R' \<and> get_TRS R' = R"
    by (transfer, auto)
  thus ?thesis unfolding mstep_rewrite'_def by auto
qed

(*copied and adapted from parallel closed*)
fun is_development_closed_cp where
  "is_development_closed_cp R _ (False, s, t) = is_mstep R s t"
| "is_development_closed_cp R (Some n) (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (reachable_terms R t n)))"
| "is_development_closed_cp R None (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (mstep_rewrite R t)))"


definition check_development_closed where
  "check_development_closed ren R n = do {
     check_wf_trs R;
     check_left_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (is_development_closed_cp R n (b, s, t))
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not (almost) development closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl ren R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not development closed\<newline>'') \<circ> showsl_trs R)"


lemma check_development_closed:
  assumes "isOK(check_development_closed ren R n)"
  shows "CR (rstep (set R))"
proof (rule mstep_closed_imp_CR)
  from assms show "left_linear_trs (set R)" 
    unfolding check_development_closed_def by auto 
  from assms have varcond:"list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) R"
    unfolding check_development_closed_def check_wf_trs_def check_varcond_no_Var_lhs list_all_iff by auto  
  then show "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
    using Critical_Pair_Closure_Impl.varcond by blast 
  show "\<And>s t. (False, s, t) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow> (s, t) \<in> mstep (set R)"
  proof -
    fix s t
    assume "(False, s, t) \<in> critical_pairs ren (set R) (set R)"
    with assms have dc:"is_development_closed_cp R n (False, s, t)" 
      unfolding check_development_closed_def by auto
    with dc have "is_mstep R s t" 
      using varcond is_development_closed_cp.simps by blast
    then show "(s, t) \<in> mstep (set R)" using is_mstep[OF varcond] by auto
  qed
  show "\<And>s t. (True, s, t) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow> \<exists>v. (s, v) \<in> mstep (set R) \<and> (t, v) \<in> (rstep (set R))\<^sup>*"
  proof -
    fix s t
    assume "(True, s, t) \<in> critical_pairs ren (set R) (set R)"
    with assms have dc:"is_development_closed_cp R n (True, s, t)" 
      unfolding check_development_closed_def by auto
    show "\<exists>v. (s, v) \<in> mstep (set R) \<and> (t, v) \<in> (rstep (set R))\<^sup>*" proof(cases n)
      case None
      with dc have "\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (mstep_rewrite R t))" 
        using varcond is_development_closed_cp.simps by auto
      then obtain v where "List.find (\<lambda>v. is_mstep R s v) (mstep_rewrite R t) = Some v"
        by force
      then have "is_mstep R s v" and "v \<in> set (mstep_rewrite R t)"
        unfolding find_Some_iff by auto
      then show ?thesis
        using is_mstep local.varcond mstep_imp_rsteps mstep_rewrite_mstep by blast 
    next
      case (Some i)
      with dc have "\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (reachable_terms R t i))" 
        using varcond is_development_closed_cp.simps by auto
      then obtain v where "List.find (\<lambda>v. is_mstep R s v) (reachable_terms R t i) = Some v"
        by force
      then have "is_mstep R s v" and "v \<in> set (reachable_terms R t i)"
        unfolding find_Some_iff by auto
      then show ?thesis
        using is_mstep local.varcond reachable_terms by blast
    qed
  qed
qed

fun is_critical_pair_closing_cp where
  "is_critical_pair_closing_cp C n (False, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. (s,v) \<in> par_rstep (set C)) (reachable_terms C t n)))"
| "is_critical_pair_closing_cp C n (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms C t n) x) (reachable_terms C s n)))"

definition check_critical_pair_closing :: "_ \<Rightarrow> ('f:: showl, _) rules \<Rightarrow> ('f, _) rules \<Rightarrow> nat \<Rightarrow> showsl check"
where
  "check_critical_pair_closing ren R C n = do {
     check_left_linear_trs R;
     check_subseteq C R <+? (\<lambda>t. showsl_lit (STR ''C not a subsystem of R''));
     check_allm (\<lambda> (b, s, t). do {
       check (is_critical_pair_closing_cp C n (b, s, t))
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl ren R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not critical pair closing\<newline>'') \<circ> showsl_trs R)"

lemma check_critical_pair_closing:
  assumes "isOK(check_critical_pair_closing ren R C n)"
    "SN (rstep (set C))"
  shows "CR (rstep (set R))"
proof (rule cpcs_sn)
  from assms show "left_linear_trs (set R)" unfolding check_critical_pair_closing_def by auto
  from assms show "SN (rstep (set C))" ..
  from assms[unfolded check_critical_pair_closing_def,simplified]
  have closed:"\<And> b p q. (b, p, q) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow> is_critical_pair_closing_cp C n (b, p, q)"
    by fast
  let ?find = "\<lambda> p v. (p,v) \<in> par_rstep (set C)" 
  show "critical_pair_closing ren (set C) (set R)" unfolding critical_pair_closing_def
  proof
    from assms show "set C \<subseteq> set R"  unfolding check_critical_pair_closing_def by auto
    { fix b p q
      assume cp:"(b, p, q) \<in> critical_pairs ren (set R) (set R)"
      have "(p, q) \<in> (rstep (set C))\<^sup>\<down>"
      proof (cases b)
        case True
        with closed[of b p q] cp
        have "\<not> Option.is_none (find (List.member (reachable_terms C q n)) (reachable_terms C p n))" by auto
        then obtain v where  "find (List.member (reachable_terms C q n)) (reachable_terms C p n) = Some v"
          by force
        then have "v \<in> set (reachable_terms C q n) \<and> v \<in> set (reachable_terms C p n)"
          unfolding find_Some_iff by (meson in_set_member nth_mem)
        then show ?thesis by (auto dest: reachable_terms)
      next
        case False
        with closed[of b p q] cp
        have "\<not> Option.is_none (find (?find p) (reachable_terms C q n))" by auto
        then obtain v where  "find (?find p) (reachable_terms C q n) = Some v"
          by force
        then have "(p,v) \<in> par_rstep (set C)" and "v \<in> set (reachable_terms C q n)"
          unfolding find_Some_iff by auto
        then show ?thesis using reachable_terms par_rstep_rsteps by blast
      qed
    }
    then show "\<forall>(b, p, q) \<in> critical_pairs ren (set R) (set R). (p, q) \<in> (rstep (set C))\<^sup>\<down>" by auto
  qed
  show "\<And>p q. (False, q, p) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow>
           \<exists>v. (q, v) \<in> par_rstep (set C) \<and> (p, v) \<in> (rstep (set C))\<^sup>*"
  proof -
    fix p q
    assume "(False, p, q) \<in> critical_pairs ren (set R) (set R)"
    with closed[of False p q]
    have "\<not> Option.is_none (find (?find p) (reachable_terms C q n))" by auto
    then obtain v where  "find (?find p) (reachable_terms C q n) = Some v"
      by force
    then have "(p,v) \<in> par_rstep (set C)" and "v \<in> set (reachable_terms C q n)"
      unfolding find_Some_iff by auto
    then show "\<exists>v. (p, v) \<in> par_rstep (set C) \<and> (q, v) \<in> (rstep (set C))\<^sup>*"
      using reachable_terms par_rstep_rsteps by blast
  qed
qed

end
